import os
import pickle as pkl

from mimeta import MIMeta
from torchcross.data.metadataset import FewShotMetaDataset, SubTaskRandomFewShotMetaDataset
from torchcross.data.task import Task

overwrite = False


def available_tasks(data_path) -> list[tuple[str, str]]:
    task_dict = MIMeta.get_available_tasks(data_path)
    return [(dataset, task) for dataset, tasks in task_dict.items() for task in tasks]


def create_few_shot_tasks(
    data_path,
    dataset_name: str,
    task_name: str,
    n_support: int,
    n_query: int,
    length: int,
) -> list[Task]:
    task_source = MIMeta(data_path, dataset_name, task_name)
    few_shot = FewShotMetaDataset(
        task_source, None, n_support, n_query, length=length, output_indices=True
    )
    print(
        f"Creating {length} few-shot tasks for: {dataset_name} {task_name} {n_support} {n_query}"
    )
    try:
        print("Length: ", len(few_shot))
    except ValueError:
        print("Length: None")
    task_list = [t for t in few_shot]
    print("Total: ", len(task_list))
    return task_list


def create_random_few_shot_tasks(
    data_path,
    dataset_name: str,
    task_name: str,
    n_support_min: int,
    n_support_max: int,
    n_query: int,
    length: int,
) -> list[Task]:
    task_source = MIMeta(data_path, dataset_name, task_name)
    few_shot = SubTaskRandomFewShotMetaDataset(
        task_source,
        None,
        n_support_samples_per_class_min=n_support_min,
        n_support_samples_per_class_max=n_support_max,
        n_query_samples_per_class=n_query,
        length=length,
        output_indices=True,
    )
    print(
        f"Creating {length} few-shot tasks for: {dataset_name} {task_name} {n_support_min}-{n_support_max} {n_query}"
    )
    try:
        print("Length: ", len(few_shot))
    except ValueError:
        print("Length: None")
    task_list = [t for t in few_shot]
    print("Total: ", len(task_list))
    return task_list


def save_few_shot_tasks(data_path, save_path=None):
    n_query = 10
    length = 100
    os.makedirs(save_path, exist_ok=True)
    # create few-shot instances for all tasks and all nshot values
    # and save them to pkl files
    for dataset, task in available_tasks(data_path):
        for n_support in [1, 2, 3, 5, 7, 10, 15, 20, 25, 30]:
            few_shot_tasks = create_few_shot_tasks(
                data_path, dataset, task, n_support, n_query, length
            )
            if few_shot_tasks is None:
                continue
            file_name = f"{dataset}_{task}_{n_support}_{n_query}_{length}.pkl"
            if os.path.exists(os.path.join(save_path, file_name)) and not overwrite:
                raise FileExistsError(
                    f"File {file_name} already exists. Set overwrite to True to overwrite."
                )
            with open(os.path.join(save_path, file_name), "wb") as f:
                pkl.dump(few_shot_tasks, f)


def save_random_few_shot_tasks(data_path, save_path=None):
    n_query = 10
    length = 1000
    n_support_min = 1
    n_support_max = 10
    os.makedirs(save_path, exist_ok=True)
    for dataset, task in available_tasks(data_path):
        few_shot_tasks = create_random_few_shot_tasks(
            data_path, dataset, task, n_support_min, n_support_max, n_query, length
        )
        if few_shot_tasks is None:
            continue
        file_name = (
            f"{dataset}_{task}_{n_support_min}-{n_support_max}_{n_query}_{length}.pkl"
        )
        if os.path.exists(os.path.join(save_path, file_name)) and not overwrite:
            raise FileExistsError(
                f"File {file_name} already exists. Set overwrite to True to overwrite."
            )
        with open(os.path.join(save_path, file_name), "wb") as f:
            pkl.dump(few_shot_tasks, f)


def main():
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("--data_path", type=str, default="data/MIMeta")
    parser.add_argument("--save_path", type=str, default="data/MIMeta_presampled")
    args = parser.parse_args()
    data_path = args.data_path
    save_path = args.save_path

    print("Available tasks:")
    print(available_tasks(data_path))
    save_few_shot_tasks(data_path, save_path)
    save_random_few_shot_tasks(data_path, save_path)


if __name__ == "__main__":
    main()



from functools import partial

import lightning.pytorch as pl
import torch
import torch.nn as _nn
import torchvision.models as _models
from torch.utils.data import DataLoader

import torchcross as tx
from mimeta import MIMeta
from torchcross.data.metadataset import FewShotMetaDataset
from torchcross.data.task_source import SubsetTaskSource, BatchedTaskSource
from torchcross.models.lightning import SimpleCrossDomainClassifier
from torchcross.utils.collate_fn import identity


def resnet18_backbone(pretrained=False):
    weights = _models.ResNet18_Weights.DEFAULT if pretrained else None
    resnet = _models.resnet18(weights=weights, num_classes=1000)
    num_features = resnet.fc.in_features
    resnet.fc = _nn.Identity()
    return resnet, num_features


def main(args):
    data_path = args.data_path
    target_dataset_name = args.target_dataset
    target_task_name = args.target_task
    num_workers = args.num_workers

    batch_size = 64

    hparams = {
        "lr": 1e-3,
    }

    # Create optimizer
    optimizer = partial(torch.optim.Adam, **hparams)

    # Create the lighting model with pre-trained resnet18 backbone
    model = SimpleCrossDomainClassifier(resnet18_backbone(pretrained=True), optimizer)

    # Create one 10-shot task from the target dataset
    target_task_source = MIMeta(data_path, target_dataset_name, target_task_name)
    target_few_shot = FewShotMetaDataset(
        target_task_source,
        collate_fn=None,
        n_support_samples_per_class=10,
        n_query_samples_per_class=10,
        filter_classes_min_samples=20,
        length=1,
        output_indices=True,
    )
    task = next(iter(target_few_shot))

    target_train_indices = [i for i, _ in task.support]
    target_test_indices = [i for i, _ in task.query]

    # Create target training and test dataloaders
    target_train_dataset = BatchedTaskSource(
        task_source=SubsetTaskSource(target_task_source, target_train_indices),
        batch_size=batch_size,
        collate_fn=tx.utils.collate_fn.stack,
        with_task_description=True,
    )
    target_test_dataset = BatchedTaskSource(
        task_source=SubsetTaskSource(target_task_source, target_test_indices),
        batch_size=batch_size,
        collate_fn=tx.utils.collate_fn.stack,
        with_task_description=True,
    )

    target_train_dataloader = DataLoader(
        target_train_dataset,
        batch_size=None,
        num_workers=num_workers,
        collate_fn=identity,
        pin_memory=True,
    )

    target_test_dataloader = DataLoader(
        target_test_dataset,
        batch_size=None,
        num_workers=num_workers,
        collate_fn=identity,
        pin_memory=True,
    )

    target_trainer = pl.Trainer(inference_mode=False, max_epochs=100)

    # Fine-tune the model on the target task
    # We are using the test set as the validation set here for simplicity
    target_trainer.fit(model, target_train_dataloader, target_test_dataloader)

    # Test the model after fine-tuning on the target task
    target_trainer.test(model, target_test_dataloader)


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("--data_path", type=str, default="data/MIMeta")
    parser.add_argument("--target_dataset", type=str, default="OCT")
    parser.add_argument("--target_task", type=str, default="disease")
    parser.add_argument("--num_workers", type=int, default=8)
    args = parser.parse_args()

    main(args)



from typing import Optional, Callable

from torchcross.data import RandomInterleaveDataset, BatchedTaskSource
from .logger import logger
from .mimeta import MIMeta


class MultiMIMetaBatchTaskSource(RandomInterleaveDataset):
    def __init__(
        self,
        mimeta_data_path: str,
        task_ids: list[tuple[str, str]],
        batch_size: int,
        shuffle: bool = True,
        drop_last: bool = False,
        collate_fn: Optional[Callable] = None,
        original_splits: list[str] | str | None = None,
        transform: Optional[Callable] = None,
    ):
        logger.info(
            f"Initializing MultiMIMetaBatchTaskSource with mimeta_data_path={mimeta_data_path}, "
            f"task_ids={task_ids}, batch_size={batch_size}, shuffle={shuffle}, "
            f"drop_last={drop_last}, collate_fn={collate_fn}, original_splits={original_splits}, "
            f"transform={transform}"
        )
        if original_splits is None or isinstance(original_splits, str):
            original_splits = [original_splits] * len(task_ids)
        else:
            assert len(original_splits) == len(task_ids)
        unbatched_task_sources = [
            MIMeta(
                data_path=mimeta_data_path,
                dataset_name=dataset_name,
                task_name=task_name,
                original_split=original_split,
                transform=transform,
            )
            for (dataset_name, task_name), original_split in zip(
                task_ids, original_splits
            )
        ]
        batched_task_sources = [
            BatchedTaskSource(
                task_source=task_source,
                batch_size=batch_size,
                shuffle=shuffle,
                drop_last=drop_last,
                collate_fn=collate_fn,
                with_task_description=True,
            )
            for task_source in unbatched_task_sources
        ]
        super().__init__(batched_task_sources)



import os
import pickle as pkl
from typing import Optional, Callable

from torchcross.data.metadataset import MetaDataset
from torchcross.data.task import Task
from torchcross.utils.collate_fn import identity
from .logger import logger
from .mimeta import MIMeta


class PickledMIMetaTaskDataset(MetaDataset):
    def __init__(
        self,
        task_lists_data_path: str,
        mimeta_data_path: str,
        dataset_name: str,
        task_name: str,
        n_support: int | tuple[int, int],
        n_query: int,
        length: int,
        collate_fn: Optional[Callable] = None,
        transform: Optional[Callable] = None,
    ):
        logger.info(
            f"Initializing PickledMIMetaTaskDataset with task_lists_data_path={task_lists_data_path}, "
            f"mimeta_data_path={mimeta_data_path}, dataset_name={dataset_name}, task_name={task_name}, "
            f"n_support={n_support}, n_query={n_query}, length={length}, "
            f"collate_fn={collate_fn}, transform={transform}"
        )
        if isinstance(n_support, int):
            file_name = f"{dataset_name}_{task_name}_{n_support}_{n_query}_{length}.pkl"
        else:
            file_name = f"{dataset_name}_{task_name}_{n_support[0]}-{n_support[1]}_{n_query}_{length}.pkl"
        file_path = os.path.join(task_lists_data_path, file_name)
        if not os.path.exists(file_path):
            raise ValueError(f"Task list file {file_path} not found.")
        with open(file_path, "rb") as f:
            self.task_list: list[Task] = pkl.load(f)
        self.task_source = MIMeta(
            data_path=mimeta_data_path,
            dataset_name=dataset_name,
            task_name=task_name,
            transform=transform,
        )
        self.collate_fn = collate_fn if collate_fn is not None else identity

    def __getitem__(self, index) -> Task:
        task: Task = self.task_list[index]
        support = [(self.task_source[i][0], l) for i, l in task.support]
        query = [(self.task_source[i][0], l) for i, l in task.query]
        support = self.collate_fn(support)
        query = self.collate_fn(query)
        return Task(support, query, task.task_target, task.classes)

    def __len__(self) -> int:
        return len(self.task_list)


class MultiPickledMIMetaTaskDataset(MetaDataset):
    def __init__(
        self,
        task_lists_data_path: str,
        mimeta_data_path: str,
        task_ids: list[tuple[str, str]],
        n_support: int | tuple[int, int],
        n_query: int,
        length: int,
        collate_fn: Optional[Callable] = None,
        transform: Optional[Callable] = None,
    ):
        logger.info(
            f"Initializing MultiPickledMIMetaTaskDataset with task_lists_data_path={task_lists_data_path}, "
            f"mimeta_data_path={mimeta_data_path}, task_ids={task_ids}, "
            f"n_support={n_support}, n_query={n_query}, length={length}, "
            f"collate_fn={collate_fn}, transform={transform}"
        )
        self.task_list: list[tuple[Task, int]] = []
        for i, (dataset_name, task_name) in enumerate(task_ids):
            if isinstance(n_support, int):
                file_name = (
                    f"{dataset_name}_{task_name}_{n_support}_{n_query}_{length}.pkl"
                )
            else:
                file_name = f"{dataset_name}_{task_name}_{n_support[0]}-{n_support[1]}_{n_query}_{length}.pkl"
            file_path = os.path.join(task_lists_data_path, file_name)
            if not os.path.exists(file_path):
                raise ValueError(f"Task list file {file_path} not found.")
            with open(file_path, "rb") as f:
                self.task_list.extend(zip(pkl.load(f), [i] * length))
        self.task_sources = [
            MIMeta(
                data_path=mimeta_data_path,
                dataset_name=dataset_name,
                task_name=task_name,
                transform=transform,
            )
            for dataset_name, task_name in task_ids
        ]
        self.collate_fn = collate_fn if collate_fn is not None else identity

    def __getitem__(self, index) -> Task:
        task, task_source_index = self.task_list[index]
        task_source = self.task_sources[task_source_index]
        support = [(task_source[i][0], l) for i, l in task.support]
        query = [(task_source[i][0], l) for i, l in task.query]
        support = self.collate_fn(support)
        query = self.collate_fn(query)
        return Task(support, query, task.task_target, task.classes)

    def __len__(self) -> int:
        return len(self.task_list)



import os

import h5py
import numpy as np
import torch
import yaml
from PIL import Image
from torchvision import transforms

from torchcross.data import TaskSource
from torchcross.data import TaskTarget
from .logger import logger


def default_transform(input_size):
    return transforms.Compose(
        [
            transforms.Resize(input_size[1:]),
            transforms.ToTensor(),
            transforms.Normalize((0.5,) * input_size[0], (0.5,) * input_size[0]),
        ]
    )


class MIMeta(TaskSource):
    _data_path: str = None
    _infos: dict[str, dict] = None
    _dataset_dir_mapping: dict[str, str] = None
    _available_tasks: dict[str, list[str]] = None

    def __init__(
        self,
        data_path,
        dataset_name,
        task_name,
        original_split=None,
        transform=None,
        use_hdf5=True,
    ):
        logger.info(
            f"Initializing MIMeta with data_path={data_path}, dataset_name={dataset_name}, "
            f"task_name={task_name}, original_split={original_split}, transform={transform}, "
            f"use_hdf5={use_hdf5}"
        )
        self.data_path = data_path
        self.dataset_name = dataset_name
        self.task_name = task_name
        self.original_split = original_split
        self.transform = transform

        available_datasets = self.get_available_datasets(data_path)
        if dataset_name not in available_datasets:
            raise ValueError(
                f"Dataset '{dataset_name}' not found. Available datasets: {available_datasets}"
            )

        dataset_subdir = self._dataset_dir_mapping[dataset_name]

        # load info.yaml file and parse task information
        info_file = os.path.join(data_path, dataset_subdir, "info.yaml")
        with open(info_file, "r") as f:
            info = yaml.load(f, Loader=yaml.FullLoader)
        task_info = next(
            (t for t in info["tasks"] if t["task_name"] == task_name), None
        )
        if task_info is None:
            raise ValueError(
                f"No task with name '{task_name}' found for dataset '{dataset_name}'"
            )
        self.classes = task_info["labels"]
        self.task_target = TaskTarget[task_info["task_target"]]
        self.input_size = info["input_size"]

        # set number of channels based on input size
        self.num_channels = self.input_size[0]

        # set transform to default if none is provided
        if self.transform is None:
            self.transform = default_transform(self.input_size)

        # set up paths and load label data
        self.image_dir = os.path.join(data_path, dataset_subdir, "images")
        self.label_file = os.path.join(
            data_path, dataset_subdir, "task_labels", f"{task_name}.npy"
        )
        self.labels: np.ndarray = np.load(self.label_file)
        if self.task_target == TaskTarget.BINARY_CLASSIFICATION:
            self.labels = self.labels[:, np.newaxis]

        self.hdf5_path = os.path.join(data_path, dataset_subdir, "images.hdf5")
        self.use_hdf5 = use_hdf5
        if self.use_hdf5:
            self.hdf5_images = h5py.File(self.hdf5_path, "r")["images"]

        if self.original_split is not None:
            split_file = os.path.join(
                data_path, dataset_subdir, "original_splits", f"{original_split}.txt"
            )
            if not os.path.exists(split_file):
                available_splits = [
                    k for k, v in info["original_splits_num_samples"].items() if v > 0
                ]
                raise ValueError(
                    f"Split '{original_split}' not found for dataset {dataset_name}. "
                    f"Available splits: {available_splits}"
                )
            with open(split_file, "r") as f:
                split_paths = [x.strip() for x in f.readlines()]
            self.split_indices = [
                int(p.split("/")[-1].split(".")[0]) for p in split_paths
            ]
            self.labels = self.labels[self.split_indices]

        self.task_identifier = f"{dataset_name}: {task_name}"

    def __getitem__(self, index):
        img_index = index if self.original_split is None else self.split_indices[index]
        if self.use_hdf5:
            image_array = self.hdf5_images[img_index, ...]
            image = Image.fromarray(image_array)
        else:
            image_path = os.path.join(self.image_dir, f"{img_index:06d}.tiff")
            image = Image.open(image_path)
        if image.mode == "RGB" and self.num_channels == 1:
            # convert RGB images to grayscale
            image = image.convert("L")
        elif image.mode == "L" and self.num_channels == 3:
            # convert grayscale images to RGB
            image = image.convert("RGB")
            # image = Image.merge("RGB", [image] * 3)
        if self.transform is not None:
            image = self.transform(image)
        label = self.labels[index]
        return image, torch.as_tensor(label)

    def __len__(self):
        return len(self.labels)

    @classmethod
    def get_available_datasets(cls, data_path: str) -> list[str]:
        if cls._dataset_dir_mapping is None or cls._data_path != data_path:
            cls._read_dataset_info(data_path)
        return list(cls._dataset_dir_mapping.keys())

    @classmethod
    def get_available_tasks(cls, data_path: str) -> dict[str, list[str]]:
        if cls._available_tasks is None or cls._data_path != data_path:
            cls._read_dataset_info(data_path)
        return cls._available_tasks

    @classmethod
    def get_info_dict(cls, data_path: str, dataset_name: str) -> dict:
        if cls._dataset_dir_mapping is None or cls._data_path != data_path:
            cls._read_dataset_info(data_path)
        if dataset_name not in cls._infos:
            raise ValueError(
                f"Dataset '{dataset_name}' not found. "
                f"Available datasets: {list(cls._infos.keys())}"
            )
        return cls._infos[dataset_name]

    @classmethod
    def _read_dataset_info(cls, data_path: str):
        cls._dataset_dir_mapping = {}
        cls._available_tasks = {}
        cls._infos = {}
        for subdir in os.listdir(data_path):
            subdir_path = os.path.join(data_path, subdir)
            if not os.path.isdir(subdir_path):
                continue
            info_path = os.path.join(subdir_path, "info.yaml")
            if not os.path.isfile(info_path):
                continue
            with open(info_path, "r") as f:
                info = yaml.load(f, Loader=yaml.FullLoader)
                cls._infos[info["name"]] = info
                cls._dataset_dir_mapping[info["name"]] = subdir
                cls._available_tasks[info["name"]] = [
                    t["task_name"] for t in info["tasks"]
                ]
        cls._data_path = data_path



from functools import partial

import lightning.pytorch as pl
import torch
import torch.nn as _nn
import torchopt
import torchvision.models as _models
from torch.utils.data import DataLoader

import torchcross as tx
from mimeta import (
    get_available_tasks,
    MultiPickledMIMetaTaskDataset, PickledMIMetaTaskDataset,
)
from torchcross.models.lightning import CrossDomainMAML
from torchcross.utils.collate_fn import identity


def resnet18_backbone(pretrained=False):
    weights = _models.ResNet18_Weights.DEFAULT if pretrained else None
    resnet = _models.resnet18(weights=weights, num_classes=1000)
    num_features = resnet.fc.in_features
    resnet.fc = _nn.Identity()
    return resnet, num_features


def main(args):
    data_path = args.data_path
    presampled_data_path = args.presampled_data_path
    target_dataset_name = args.target_dataset
    target_task_name = args.target_task
    validation_dataset_name = args.validation_dataset
    validation_task_name = args.validation_task
    num_workers = args.num_workers

    batch_size = 64

    task_dict = get_available_tasks(data_path)
    train_tasks = [
        (ds, t)
        for ds, tasks in task_dict.items()
        for t in tasks
        if ds != target_dataset_name
    ]

    # Create the cross-domain meta-dataset from pre-sampled tasks
    train_dataset = MultiPickledMIMetaTaskDataset(
        presampled_data_path,
        data_path,
        train_tasks,
        n_support=(1, 10),
        n_query=10,
        length=1000,
        collate_fn=tx.utils.collate_fn.stack,
    )
    val_dataset = PickledMIMetaTaskDataset(
        presampled_data_path,
        data_path,
        validation_dataset_name,
        validation_task_name,
        n_support=5,
        n_query=20,
        length=400,
        collate_fn=tx.utils.collate_fn.stack,
    )

    train_dataloader = DataLoader(
        train_dataset,
        batch_size=2,
        num_workers=num_workers,
        collate_fn=identity,
        pin_memory=True,
    )
    val_dataloader = DataLoader(
        val_dataset,
        batch_size=2,
        num_workers=num_workers,
        collate_fn=identity,
        pin_memory=True,
    )

    # Create optimizer
    outer_optimizer = partial(torch.optim.Adam, lr=0.001)
    inner_optimizer = partial(torchopt.MetaSGD, lr=0.1)
    eval_inner_optimizer = partial(torch.optim.SGD, lr=0.1)
    num_inner_steps = 4
    eval_num_inner_steps = 32

    # Create the lighting model with pre-trained resnet18 backbone
    model = CrossDomainMAML(resnet18_backbone(pretrained=True),
                            outer_optimizer,
                            inner_optimizer,
                            eval_inner_optimizer,
                            num_inner_steps,
                            eval_num_inner_steps)



    # Create the lightning trainer
    trainer = pl.Trainer(
        inference_mode=False,
        max_epochs=1,
        check_val_every_n_epoch=1,
        val_check_interval=1000,
        limit_val_batches=100,
    )

    # Pre-train the model on all the tasks in MIMeta except the target task
    trainer.fit(model, train_dataloader, val_dataloader)

    # Save the model
    # Uncomment the following line to save the pretrained model
    # torch.save(model.state_dict(), "model.pt")

    # Create the test dataloader
    test_dataset = PickledMIMetaTaskDataset(
        presampled_data_path,
        data_path,
        target_dataset_name,
        target_task_name,
        n_support=5,
        n_query=20,
        length=400,
        collate_fn=tx.utils.collate_fn.stack,
    )
    test_dataloader = DataLoader(
        test_dataset,
        batch_size=2,
        num_workers=num_workers,
        collate_fn=identity,
        pin_memory=True,
    )

    # Meta-test the model on the target task
    trainer.test(model, test_dataloader)


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("--data_path", type=str, default="data/MIMeta")
    parser.add_argument(
        "--presampled_data_path", type=str, default="data/MIMeta_presampled"
    )
    parser.add_argument("--target_dataset", type=str, default="OCT")
    parser.add_argument("--target_task", type=str, default="disease")
    parser.add_argument("--validation_dataset", type=str, default="OCT")
    parser.add_argument("--validation_task", type=str, default="disease")
    parser.add_argument("--num_workers", type=int, default=8)
    args = parser.parse_args()

    main(args)



from .mimeta import MIMeta
from .batches import MultiMIMetaBatchTaskSource
from .tasks import PickledMIMetaTaskDataset, MultiPickledMIMetaTaskDataset

get_available_datasets = MIMeta.get_available_datasets
get_available_tasks = MIMeta.get_available_tasks

__all__ = [
    "MIMeta",
    "MultiMIMetaBatchTaskSource",
    "PickledMIMetaTaskDataset",
    "MultiPickledMIMetaTaskDataset",
    "get_available_datasets",
    "get_available_tasks",
]
